@@ -37,6 +37,7 @@
 #include "cfg.h"
 #include "cfg_grid.h"
 #include "cfg_node.h"
+#include "cfg_buf.h"
 
 /* internal specification scanner/parser context */
 typedef struct {
@@ -46,6 +47,7 @@ typedef struct {
     cfg_t        *cfg;        /* the configuration object */
     cfg_node_t   *node;       /* top-level/root/result channel */
     cfg_rc_t      rv;         /* return value */
+    cfg_buf_t    *buf;        /* temporary buffer */
     char         *err_buf;    /* error buffer pointer */
     size_t        err_len;    /* error buffer length */
     void         *yyscan;     /* Flex scanner context */
