@@ -133,7 +133,7 @@ extern int yylex(/*YYSTYPE*/ void *lvalp,
 /* generate verbose error messages and remember them inside the context */
 #undef  yyerror
 #define yyerror(msg) \
-    cfg_syn_error(CTX, CFG_ERR_SYN, &yylloc, msg)
+    cfg_syn_error(CTX, CFG_ERR_SYN, &yylloc, "%s", msg)
 
 /* scanner state transition */
 extern void cfg_syn_scan_push(cfg_syn_ctx_t *, const char *state);
@@ -1162,7 +1162,7 @@ yyreduce:
         cfg_node_t *n;
         cfg_node_create(CTX->cfg, &n);
         cfg_node_set(CTX->cfg, n, CFG_NODE_ATTR_TYPE, CFG_NODE_TYPE_ARG);
-        cfg_node_set(CTX->cfg, n, CFG_NODE_ATTR_TOKEN, yyvsp[0].cpString);
+        cfg_node_set(CTX->cfg, n, CFG_NODE_ATTR_TOKEN|CFG_ATTR_GIFT, yyvsp[0].cpString);
         yyval.npNode = n;
       ;}
     break;
